//--------------------------------------------------------------------------
// Program to pull the information out of various types of EXIF digital 
// camera files and show it in a reasonably consistent way
//
// This module parses the very complicated exif structures.
//
// Matthias Wandel,  Dec 1999 - Dec 2002 
//--------------------------------------------------------------------------
#ifdef WIN32
#pragma comment(lib, "c:\\TopfieldDisk\\DataFiles\\SDK\\TAP_SDK_DLL")
#include "TAP_Emulations.h"
#endif

#include "tap.h"
#include "translations.h"

//#include <stdio.h>
//#include <stdlib.h>
//#include <math.h>
//#include <string.h>
//#include <time.h>
//#include <errno.h>
//#include <ctype.h>

/*#ifdef _WIN32
    #include <sys/utime.h>
#else
    #include <utime.h>
    #include <sys/types.h>
    #include <unistd.h>
    #include <errno.h>
    #include <limits.h>
#endif*/

#include "jheadinc.c"

static unsigned char * DirWithThumbnailPtrs;
//static double FocalplaneXRes;
//static double FocalplaneUnits;
static int ExifImageWidth;
static int MotorolaOrder = 0;

// for fixing the rotation.
static void * OrientationPtr;
static int    OrientationNumFormat; 

typedef struct {
    unsigned short Tag;
    char * Desc;
}TagTable_t;


//--------------------------------------------------------------------------
// Table of Jpeg encoding process names
static const TagTable_t ProcessTable[] = {
    { M_SOF0,   "Baseline"},
    { M_SOF1,   "Extended sequential"},
    { M_SOF2,   "Progressive"},
    { M_SOF3,   "Lossless"},
    { M_SOF5,   "Differential sequential"},
    { M_SOF6,   "Differential progressive"},
    { M_SOF7,   "Differential lossless"},
    { M_SOF9,   "Extended sequential, arithmetic coding"},
    { M_SOF10,  "Progressive, arithmetic coding"},
    { M_SOF11,  "Lossless, arithmetic coding"},
    { M_SOF13,  "Differential sequential, arithmetic coding"},
    { M_SOF14,  "Differential progressive, arithmetic coding"},
    { M_SOF15,  "Differential lossless, arithmetic coding"},
};

#define PROGRESSIVE 194

#define PROCESS_TABLE_SIZE  (sizeof(ProcessTable) / sizeof(TagTable_t))

// 1 - "The 0th row is at the visual top of the image,    and the 0th column is the visual left-hand side."
// 2 - "The 0th row is at the visual top of the image,    and the 0th column is the visual right-hand side."
// 3 - "The 0th row is at the visual bottom of the image, and the 0th column is the visual right-hand side."
// 4 - "The 0th row is at the visual bottom of the image, and the 0th column is the visual left-hand side."

// 5 - "The 0th row is the visual left-hand side of of the image,  and the 0th column is the visual top."
// 6 - "The 0th row is the visual right-hand side of of the image, and the 0th column is the visual top."
// 7 - "The 0th row is the visual right-hand side of of the image, and the 0th column is the visual bottom."
// 8 - "The 0th row is the visual left-hand side of of the image,  and the 0th column is the visual bottom."

// Note: The descriptions here are the same as the name of the command line
// option to pass to jpegtran to right the image

static const char OrientTab[9][20] = {
    "Undefined",
    "Normal",           // 1
    "flip horizontal",  // left right reversed mirror
    "rotate 180",       // 3
    "flip vertical",    // upside down mirror
    "transpose",        // Flipped about top-left <--> bottom-right axis.
    "rotate 90",        // rotate 90 cw to right it.
    "transverse",       // flipped about top-right <--> bottom-left axis
    "rotate 270",       // rotate 270 to right it.
};

const int BytesPerFormat[] = {0,1,1,2,4,8,1,1,2,4,8,4,8};

//--------------------------------------------------------------------------
// Describes tag values

#define TAG_MAKE               0x010F
#define TAG_MODEL              0x0110
#define TAG_ORIENTATION        0x0112
#define TAG_DATETIME           0x0132
#define TAG_THUMBNAIL_OFFSET   0x0201
#define TAG_THUMBNAIL_LENGTH   0x0202
#define TAG_EXPOSURETIME       0x829A
#define TAG_FNUMBER            0x829D
#define TAG_EXIF_OFFSET        0x8769
#define TAG_EXPOSURE_PROGRAM   0x8822
#define TAG_GPSINFO            0x8825
#define TAG_ISO_EQUIVALENT     0x8827
#define TAG_DATETIME_ORIGINAL  0x9003
#define TAG_DATETIME_DIGITIZED 0x9004
#define TAG_SHUTTERSPEED       0x9201
#define TAG_APERTURE           0x9202
#define TAG_EXPOSURE_BIAS      0x9204
#define TAG_MAXAPERTURE        0x9205
#define TAG_SUBJECT_DISTANCE   0x9206
#define TAG_METERING_MODE      0x9207
#define TAG_LIGHT_SOURCE       0x9208
#define TAG_FLASH              0x9209
#define TAG_FOCALLENGTH        0x920A
#define TAG_MAKER_NOTE         0x927C
#define TAG_USERCOMMENT        0x9286
#define TAG_EXIF_IMAGEWIDTH    0xa002
#define TAG_EXIF_IMAGELENGTH   0xa003
#define TAG_INTEROP_OFFSET     0xa005
#define TAG_FOCALPLANEXRES     0xa20E
#define TAG_FOCALPLANEUNITS    0xa210
#define TAG_EXPOSURE_INDEX     0xa215
#define TAG_EXPOSURE_MODE      0xa402
#define TAG_WHITEBALANCE       0xa403
#define TAG_DIGITALZOOMRATIO   0xA404
#define TAG_FOCALLENGTH_35MM   0xa405

static const TagTable_t TagTable[] = {
  { 0x001,   "InteropIndex"},
  { 0x002,   "InteropVersion"},
  { 0x100,   "ImageWidth"},
  { 0x101,   "ImageLength"},
  { 0x102,   "BitsPerSample"},
  { 0x103,   "Compression"},
  { 0x106,   "PhotometricInterpretation"},
  { 0x10A,   "FillOrder"},
  { 0x10D,   "DocumentName"},
  { 0x10E,   "ImageDescription"},
  { 0x10F,   "Make"},
  { 0x110,   "Model"},
  { 0x111,   "StripOffsets"},
  { 0x112,   "Orientation"},
  { 0x115,   "SamplesPerPixel"},
  { 0x116,   "RowsPerStrip"},
  { 0x117,   "StripByteCounts"},
  { 0x11A,   "XResolution"},
  { 0x11B,   "YResolution"},
  { 0x11C,   "PlanarConfiguration"},
  { 0x128,   "ResolutionUnit"},
  { 0x12D,   "TransferFunction"},
  { 0x131,   "Software"},
  { 0x132,   "DateTime"},
  { 0x13B,   "Artist"},
  { 0x13E,   "WhitePoint"},
  { 0x13F,   "PrimaryChromaticities"},
  { 0x156,   "TransferRange"},
  { 0x200,   "JPEGProc"},
  { 0x201,   "ThumbnailOffset"},
  { 0x202,   "ThumbnailLength"},
  { 0x211,   "YCbCrCoefficients"},
  { 0x212,   "YCbCrSubSampling"},
  { 0x213,   "YCbCrPositioning"},
  { 0x214,   "ReferenceBlackWhite"},
  { 0x1001,  "RelatedImageWidth"},
  { 0x1002,  "RelatedImageLength"},
  { 0x828D,  "CFARepeatPatternDim"},
  { 0x828E,  "CFAPattern"},
  { 0x828F,  "BatteryLevel"},
  { 0x8298,  "Copyright"},
  { 0x829A,  "ExposureTime"},
  { 0x829D,  "FNumber"},
  { 0x83BB,  "IPTC/NAA"},
  { 0x8769,  "ExifOffset"},
  { 0x8773,  "InterColorProfile"},
  { 0x8822,  "ExposureProgram"},
  { 0x8824,  "SpectralSensitivity"},
  { 0x8825,  "GPS Dir offset"},
  { 0x8827,  "ISOSpeedRatings"},
  { 0x8828,  "OECF"},
  { 0x9000,  "ExifVersion"},
  { 0x9003,  "DateTimeOriginal"},
  { 0x9004,  "DateTimeDigitized"},
  { 0x9101,  "ComponentsConfiguration"},
  { 0x9102,  "CompressedBitsPerPixel"},
  { 0x9201,  "ShutterSpeedValue"},
  { 0x9202,  "ApertureValue"},
  { 0x9203,  "BrightnessValue"},
  { 0x9204,  "ExposureBiasValue"},
  { 0x9205,  "MaxApertureValue"},
  { 0x9206,  "SubjectDistance"},
  { 0x9207,  "MeteringMode"},
  { 0x9208,  "LightSource"},
  { 0x9209,  "Flash"},
  { 0x920A,  "FocalLength"},
  { 0x927C,  "MakerNote"},
  { 0x9286,  "UserComment"},
  { 0x9290,  "SubSecTime"},
  { 0x9291,  "SubSecTimeOriginal"},
  { 0x9292,  "SubSecTimeDigitized"},
  { 0xA000,  "FlashPixVersion"},
  { 0xA001,  "ColorSpace"},
  { 0xA002,  "ExifImageWidth"},
  { 0xA003,  "ExifImageLength"},
  { 0xA004,  "RelatedAudioFile"},
  { 0xA005,  "InteroperabilityOffset"},
  { 0xA20B,  "FlashEnergy"},              
  { 0xA20C,  "SpatialFrequencyResponse"}, 
  { 0xA20E,  "FocalPlaneXResolution"},    
  { 0xA20F,  "FocalPlaneYResolution"},    
  { 0xA210,  "FocalPlaneResolutionUnit"}, 
  { 0xA214,  "SubjectLocation"},          
  { 0xA215,  "ExposureIndex"},            
  { 0xA217,  "SensingMethod"},            
  { 0xA300,  "FileSource"},
  { 0xA301,  "SceneType"},
  { 0xA301,  "CFA Pattern"},
  { 0xA401,  "CustomRendered"},
  { 0xA402,  "ExposureMode"},
  { 0xA403,  "WhiteBalance"},
  { 0xA404,  "DigitalZoomRatio"},
  { 0xA405,  "FocalLengthIn35mmFilm"},
  { 0xA406,  "SceneCaptureType"},
  { 0xA407,  "GainControl"},
  { 0xA408,  "Contrast"},
  { 0xA409,  "Saturation"},
  { 0xA40a,  "Sharpness"},
  { 0xA40c,  "SubjectDistanceRange"},
} ;

#define TAG_TABLE_SIZE  (sizeof(TagTable) / sizeof(TagTable_t))

struct exif
{
	char heading[20];
	char detail[100];
} photo_exif[32];
int no_of_details;

//--------------------------------------------------------------------------
// Convert a 16 bit unsigned value to file's native byte order
//--------------------------------------------------------------------------
static void Put16u(void * Short, unsigned short PutValue)
{
    if (MotorolaOrder){
        ((uchar *)Short)[0] = (uchar)(PutValue>>8);
        ((uchar *)Short)[1] = (uchar)PutValue;
    }else{
        ((uchar *)Short)[0] = (uchar)PutValue;
        ((uchar *)Short)[1] = (uchar)(PutValue>>8);
    }
}

//--------------------------------------------------------------------------
// Convert a 16 bit unsigned value from file's native byte order
//--------------------------------------------------------------------------
int Get16u(void * Short)
{
    if (MotorolaOrder){
        return (((uchar *)Short)[0] << 8) | ((uchar *)Short)[1];
    }else{
        return (((uchar *)Short)[1] << 8) | ((uchar *)Short)[0];
    }
}

//--------------------------------------------------------------------------
// Convert a 32 bit signed value from file's native byte order
//--------------------------------------------------------------------------
int Get32s(void * Long)
{
    if (MotorolaOrder){
        return  ((( char *)Long)[0] << 24) | (((uchar *)Long)[1] << 16)
              | (((uchar *)Long)[2] << 8 ) | (((uchar *)Long)[3] << 0 );
    }else{
        return  ((( char *)Long)[3] << 24) | (((uchar *)Long)[2] << 16)
              | (((uchar *)Long)[1] << 8 ) | (((uchar *)Long)[0] << 0 );
    }
}

//--------------------------------------------------------------------------
// Convert a 32 bit unsigned value to file's native byte order
//--------------------------------------------------------------------------
void Put32u(void * Value, unsigned PutValue)
{
    if (MotorolaOrder){
        ((uchar *)Value)[0] = (uchar)(PutValue>>24);
        ((uchar *)Value)[1] = (uchar)(PutValue>>16);
        ((uchar *)Value)[2] = (uchar)(PutValue>>8);
        ((uchar *)Value)[3] = (uchar)PutValue;
    }else{
        ((uchar *)Value)[0] = (uchar)PutValue;
        ((uchar *)Value)[1] = (uchar)(PutValue>>8);
        ((uchar *)Value)[2] = (uchar)(PutValue>>16);
        ((uchar *)Value)[3] = (uchar)(PutValue>>24);
    }
}

//--------------------------------------------------------------------------
// Convert a 32 bit unsigned value from file's native byte order
//--------------------------------------------------------------------------
unsigned Get32u(void * Long)
{
    return (unsigned)Get32s(Long) & 0xffffffff;
}

//--------------------------------------------------------------------------
// Display a number as one of its many formats
//--------------------------------------------------------------------------
/*void PrintFormatNumber(void * ValuePtr, int Format, int ByteCount)
{
    int s,n;

    for(n=0;n<16;n++){
        switch(Format){
            case FMT_SBYTE:
            case FMT_BYTE:      printf("%02x",*(uchar *)ValuePtr); s=1;  break;
            case FMT_USHORT:    printf("%d",Get16u(ValuePtr)); s=2;      break;
            case FMT_ULONG:     
            case FMT_SLONG:     printf("%d",Get32s(ValuePtr)); s=4;      break;
            case FMT_SSHORT:    printf("%hd",(signed short)Get16u(ValuePtr)); s=2; break;
            case FMT_URATIONAL:
            case FMT_SRATIONAL: 
               printf("%d/%d",Get32s(ValuePtr), Get32s(4+(char *)ValuePtr)); 
               s = 8;
               break;

            case FMT_SINGLE:    printf("%f",(double)*(float *)ValuePtr); s=8; break;
            case FMT_DOUBLE:    printf("%f",*(double *)ValuePtr);        s=8; break;
            default: 
                printf("Unknown format %d:", Format);
                return;
        }
        ByteCount -= s;
        if (ByteCount <= 0) break;
        printf(", ");
        ((char *)ValuePtr) += s;
    }
    if (n >= 16) printf("...");
}*/

int getRational(void * ValuePtr, bool numerator)
{
	if (numerator)
		return Get32s(ValuePtr);

	return Get32s(4+(char *)ValuePtr);
}

//--------------------------------------------------------------------------
// Evaluate number, be it int, rational, or float from directory.
//--------------------------------------------------------------------------
int ConvertAnyFormat(void * ValuePtr, int Format)
{
    int Value = 0;

    switch(Format){
        case FMT_SBYTE:     Value = *(signed char *)ValuePtr;  break;
        case FMT_BYTE:      Value = *(uchar *)ValuePtr;        break;

        case FMT_USHORT:    Value = Get16u(ValuePtr);          break;
        case FMT_ULONG:     Value = Get32u(ValuePtr);          break;

        case FMT_URATIONAL:
        case FMT_SRATIONAL: // see getRational instead, since Toppy doesn't like floating points
            {
                /*int Num,Den;
                Num = Get32s(ValuePtr);
                Den = Get32s(4+(char *)ValuePtr);
                if (Den == 0){
                    Value = 0;
                }else{
                    Value = (double)Num/Den;
                }*/
                break;
            }

        case FMT_SSHORT:    Value = (signed short)Get16u(ValuePtr);  break;
        case FMT_SLONG:     Value = Get32s(ValuePtr);                break;

        // Not sure if this is correct (never seen float used in Exif format)
        //case FMT_SINGLE:    Value = (double)*(float *)ValuePtr;      break;
        //case FMT_DOUBLE:    Value = *(double *)ValuePtr;             break;

		// ignore floats and doubles, since they do not seem to be likely to appear and Toppy doesn't handle them anyway
		// if needed, then will need to get value as a char and search for decimal point to store as int and dec parts
		case FMT_SINGLE:    Value = 0;      break;
        case FMT_DOUBLE:    Value = 0;      break;

        default:
            printf("Illegal format code %d",Format);
    }
    return Value;
}

//--------------------------------------------------------------------------
// Process one of the nested EXIF directories.
//--------------------------------------------------------------------------
static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase, 
        unsigned ExifLength, int NestingLevel)
{
    int de;
    int a;
    int NumDirEntries;
    unsigned ThumbnailOffset = 0;
    unsigned ThumbnailSize = 0;
    char IndentString[25];

    if (NestingLevel > 4){
        printf("Maximum directory nesting exceeded (corrupt exif header)");
        return;
    }

    memset(IndentString, ' ', 25);
    IndentString[NestingLevel * 4] = '\0';


    NumDirEntries = Get16u(DirStart);
    #define DIR_ENTRY_ADDR(Start, Entry) (Start+2+12*(Entry))

    {
        unsigned char * DirEnd;
        DirEnd = DIR_ENTRY_ADDR(DirStart, NumDirEntries);
        if (DirEnd+4 > (OffsetBase+ExifLength)){
            if (DirEnd+2 == OffsetBase+ExifLength || DirEnd == OffsetBase+ExifLength){
                // Version 1.3 of jhead would truncate a bit too much.
                // This also caught later on as well.
            }else{
                printf("Illegally sized directory");
                return;
            }
        }
        if (DumpExifMap){
            printf("Map: %05d-%05d: Directory\n",DirStart-OffsetBase, DirEnd+4-OffsetBase);
        }


    }

    if (ShowTags){
        printf("(dir has %d entries)\n",NumDirEntries);
    }

    for (de=0;de<NumDirEntries;de++){
        int Tag, Format, Components;
        unsigned char * ValuePtr;
        int ByteCount;
        unsigned char * DirEntry;
        DirEntry = DIR_ENTRY_ADDR(DirStart, de);

        Tag = Get16u(DirEntry);
        Format = Get16u(DirEntry+2);
        Components = Get32u(DirEntry+4);

        if ((Format-1) >= NUM_FORMATS) {
            // (-1) catches illegal zero case as unsigned underflows to positive large.
            printf("Illegal number format %d for tag %04x", Format, Tag);
            continue;
        }

        ByteCount = Components * BytesPerFormat[Format];

        if (ByteCount > 4){
            unsigned OffsetVal;
            OffsetVal = Get32u(DirEntry+8);
            // If its bigger than 4 bytes, the dir entry contains an offset.
            if (OffsetVal+ByteCount > ExifLength){
                // Bogus pointer offset and / or bytecount value
                printf("Illegal value pointer for tag %04x", Tag);
                continue;
            }
            ValuePtr = OffsetBase+OffsetVal;

            if (OffsetVal > ImageInfo.LargestExifOffset){
                ImageInfo.LargestExifOffset = OffsetVal;
            }

            if (DumpExifMap){
                printf("Map: %05d-%05d:   Data for tag %04x\n",OffsetVal, OffsetVal+ByteCount, Tag);
            }
        }else{
            // 4 bytes or less and value is in the dir entry itself
            ValuePtr = DirEntry+8;
        }

        if (Tag == TAG_MAKER_NOTE){
            if (ShowTags){
                printf("%s    Maker note: ",IndentString);
            }
            ProcessMakerNote(ValuePtr, ByteCount, OffsetBase, ExifLength);
            continue;
        }

        if (ShowTags){
            // Show tag name
            for (a=0;;a++){
                if (a >= TAG_TABLE_SIZE){
                    printf(IndentString);
                    printf("    Unknown Tag %04x Value = ", Tag);
                    break;
                }
                if (TagTable[a].Tag == Tag){
                    printf(IndentString);
                    printf("    %s = ",TagTable[a].Desc);
                    break;
                }
            }

            // Show tag value.
            switch(Format){
                case FMT_BYTE:
                    if(ByteCount>1){
                        printf("%.*ls\n", ByteCount/2, (wchar_t *)ValuePtr);
                    }else{
                        //PrintFormatNumber(ValuePtr, Format, ByteCount);
                        printf("\n");
                    }
                    break;

                case FMT_UNDEFINED:
                    // Undefined is typically an ascii string.

                case FMT_STRING:
                    // String arrays printed without function call (different from int arrays)
                    {
                        int NoPrint = 0;
                        printf("\"");
                        for (a=0;a<ByteCount;a++){
                            if (ValuePtr[a] >= 32){
                                //putchar(ValuePtr[a]);
                                NoPrint = 0;
                            }else{
                                // Avoiding indicating too many unprintable characters of proprietary
                                // bits of binary information this program may not know how to parse.
                                if (!NoPrint && a != ByteCount-1){
                                    //putchar('?');
                                    NoPrint = 1;
                                }
                            }
                        }
                        printf("\"\n");
                    }
                    break;

                default:
                    // Handle arrays of numbers later (will there ever be?)
                    //PrintFormatNumber(ValuePtr, Format, ByteCount);
                    printf("\n");
            }
        }

        // Extract useful components of tag
        switch(Tag){

            case TAG_MAKE:
                strncpy(ImageInfo.CameraMake, (char *)ValuePtr, ByteCount < 31 ? ByteCount : 31);
                break;

            case TAG_MODEL:
                strncpy(ImageInfo.CameraModel, (char *)ValuePtr, ByteCount < 39 ? ByteCount : 39);
                break;

            case TAG_DATETIME_ORIGINAL:
                // If we get a DATETIME_ORIGINAL, we use that one.
                strncpy(ImageInfo.DateTime, (char *)ValuePtr, 19);
                // Fallthru...

            case TAG_DATETIME_DIGITIZED:
            case TAG_DATETIME:
                if (!isdigit(ImageInfo.DateTime[0])){
                    // If we don't already have a DATETIME_ORIGINAL, use whatever
                    // time fields we may have.
                    strncpy(ImageInfo.DateTime, (char *)ValuePtr, 19);
                }

                if (ImageInfo.numDateTimeTags >= MAX_DATE_COPIES){
                    printf("More than %d date fields!  This is nuts", MAX_DATE_COPIES);
                    break;
                }
                ImageInfo.DateTimeOffsets[ImageInfo.numDateTimeTags++] = 
                    (char *)ValuePtr - (char *)OffsetBase;
                break;


            case TAG_USERCOMMENT:
                // Olympus has this padded with trailing spaces.  Remove these first.
                for (a=ByteCount;;){
                    a--;
                    if ((ValuePtr)[a] == ' '){
                        (ValuePtr)[a] = '\0';
                    }else{
                        break;
                    }
                    if (a == 0) break;
                }

                // Copy the comment
                if (memcmp(ValuePtr, "ASCII",5) == 0){
                    for (a=5;a<10;a++){
                        int c;
                        c = (ValuePtr)[a];
                        if (c != '\0' && c != ' '){
                            strncpy(ImageInfo.Comments, (char *)ValuePtr+a, 199);
                            break;
                        }
                    }
                    
                }else{
                    strncpy(ImageInfo.Comments, (char *)ValuePtr, 199);
                }
                break;

            case TAG_FNUMBER:
                // Simplest way of expressing aperture, so I trust it the most.
                // (overwrite previously computd value if there is one)
				if (Format == FMT_URATIONAL || Format == FMT_SRATIONAL)
				{
					ImageInfo.ApertureFNumber_Num = getRational(ValuePtr, TRUE);
					ImageInfo.ApertureFNumber_Den = getRational(ValuePtr, FALSE);
				}
				//ImageInfo.ApertureFNumber = (float)ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_APERTURE:
            case TAG_MAXAPERTURE:
                // More relevant info always comes earlier, so only use this field if we don't 
                // have appropriate aperture information yet.
                //if (ImageInfo.ApertureFNumber == 0){
                //    ImageInfo.ApertureFNumber 
                //        = (float)exp(ConvertAnyFormat(ValuePtr, Format)*log(2)*0.5);
                //}
                break;

            case TAG_FOCALLENGTH:
                // Nice digital cameras actually save the focal length as a function
                // of how farthey are zoomed in.
				if (Format == FMT_URATIONAL || Format == FMT_SRATIONAL)
				{
					ImageInfo.FocalLength_Num = getRational(ValuePtr, TRUE);
					ImageInfo.FocalLength_Den = getRational(ValuePtr, FALSE);
				}
                //ImageInfo.FocalLength = (float)ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_SUBJECT_DISTANCE:
                // Inidcates the distacne the autofocus camera is focused to.
                // Tends to be less accurate as distance increases.
				if (Format == FMT_URATIONAL || Format == FMT_SRATIONAL)
				{
					ImageInfo.Distance_Num = getRational(ValuePtr, TRUE);
					ImageInfo.Distance_Den = getRational(ValuePtr, FALSE);
				}
                //ImageInfo.Distance = (float)ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_EXPOSURETIME:
                // Simplest way of expressing exposure time, so I trust it most.
                // (overwrite previously computd value if there is one)
				if (Format == FMT_URATIONAL || Format == FMT_SRATIONAL)
				{
					ImageInfo.ExposureTime_Num = getRational(ValuePtr, TRUE);
					ImageInfo.ExposureTime_Den = getRational(ValuePtr, FALSE);
				}
                //ImageInfo.ExposureTime = (float)ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_SHUTTERSPEED:
                // More complicated way of expressing exposure time, so only use
                // this value if we don't already have it from somewhere else.
                //if (ImageInfo.ExposureTime == 0){
                //    ImageInfo.ExposureTime 
                //        = (float)(1/exp(ConvertAnyFormat(ValuePtr, Format)*log(2)));
                //}
                break;


            case TAG_FLASH:
                ImageInfo.FlashUsed=(int)ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_ORIENTATION:
                if (OrientationPtr){
                    // Already have orientation.  The second orientation is likely
                    // to be that of the thumbnail image.
                    break;
                }
                ImageInfo.Orientation = (int)ConvertAnyFormat(ValuePtr, Format);
                OrientationPtr = ValuePtr;
                OrientationNumFormat = Format;
                if (ImageInfo.Orientation < 0 || ImageInfo.Orientation > 8){
                    printf("Undefined rotation value %d", ImageInfo.Orientation);
                    ImageInfo.Orientation = 0;
                }
                break;

            case TAG_EXIF_IMAGELENGTH:
            case TAG_EXIF_IMAGEWIDTH:
                // Use largest of height and width to deal with images that have been
                // rotated to portrait format.
                a = (int)ConvertAnyFormat(ValuePtr, Format);
                if (ExifImageWidth < a) ExifImageWidth = a;
                break;

            case TAG_FOCALPLANEXRES:
                //FocalplaneXRes = ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_FOCALPLANEUNITS:
                /*switch((int)ConvertAnyFormat(ValuePtr, Format)){
                    case 1: FocalplaneUnits = 25.4; break; // inch
                    case 2: 
                        // According to the information I was using, 2 means meters.
                        // But looking at the Cannon powershot's files, inches is the only
                        // sensible value.
                        FocalplaneUnits = 25.4;
                        break;

                    case 3: FocalplaneUnits = 10;   break;  // centimeter
                    case 4: FocalplaneUnits = 1;    break;  // millimeter
                    case 5: FocalplaneUnits = .001; break;  // micrometer
                }*/
                break;

            case TAG_EXPOSURE_BIAS:
				if (Format == FMT_URATIONAL || Format == FMT_SRATIONAL)
				{
					ImageInfo.ExposureBias_Num = getRational(ValuePtr, TRUE);
					ImageInfo.ExposureBias_Den = getRational(ValuePtr, FALSE);
				}
                //ImageInfo.ExposureBias = (float)ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_WHITEBALANCE:
                ImageInfo.Whitebalance = (int)ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_LIGHT_SOURCE:
                ImageInfo.LightSource = (int)ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_METERING_MODE:
                ImageInfo.MeteringMode = (int)ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_EXPOSURE_PROGRAM:
                ImageInfo.ExposureProgram = (int)ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_EXPOSURE_INDEX:
                if (ImageInfo.ISOequivalent == 0){
                    // Exposure idx and ISO equivalent are often used interchangeably,
                    // so we will do the same in jhead.
                    // http://photography.about.com/library/glossary/bldef_ei.htm
                    ImageInfo.ISOequivalent = (int)ConvertAnyFormat(ValuePtr, Format);
                }
                break;

            case TAG_EXPOSURE_MODE:
                ImageInfo.ExposureMode = (int)ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_ISO_EQUIVALENT:
                ImageInfo.ISOequivalent = (int)ConvertAnyFormat(ValuePtr, Format);
                if ( ImageInfo.ISOequivalent < 50 ){
                    // Fixes strange encoding on some older digicams.
                    ImageInfo.ISOequivalent *= 200;
                }
                break;

            case TAG_DIGITALZOOMRATIO:
				if (Format == FMT_URATIONAL || Format == FMT_SRATIONAL)
				{
					ImageInfo.DigitalZoomRatio_Num = getRational(ValuePtr, TRUE);
					ImageInfo.DigitalZoomRatio_Den = getRational(ValuePtr, FALSE);
				}
                //ImageInfo.DigitalZoomRatio = (float)ConvertAnyFormat(ValuePtr, Format);
                break;

            case TAG_THUMBNAIL_OFFSET:
                ThumbnailOffset = (unsigned)ConvertAnyFormat(ValuePtr, Format);
                DirWithThumbnailPtrs = DirStart;
                break;

            case TAG_THUMBNAIL_LENGTH:
                ThumbnailSize = (unsigned)ConvertAnyFormat(ValuePtr, Format);
                ImageInfo.ThumbnailSizeOffset = ValuePtr-OffsetBase;
                break;

            case TAG_EXIF_OFFSET:
                if (ShowTags) printf("%s    Exif Dir:",IndentString);

            case TAG_INTEROP_OFFSET:
                if (Tag == TAG_INTEROP_OFFSET && ShowTags) printf("%s    Interop Dir:",IndentString);
                {
                    unsigned char * SubdirStart;
                    SubdirStart = OffsetBase + Get32u(ValuePtr);
                    if (SubdirStart < OffsetBase || SubdirStart > OffsetBase+ExifLength){
                        printf("Illegal exif or interop ofset directory link");
                    }else{
                        ProcessExifDir(SubdirStart, OffsetBase, ExifLength, NestingLevel+1);
                    }
                    continue;
                }
                break;

            case TAG_GPSINFO:
                if (ShowTags) printf("%s    GPS info dir:",IndentString);
                {
                    unsigned char * SubdirStart;
                    SubdirStart = OffsetBase + Get32u(ValuePtr);
                    if (SubdirStart < OffsetBase || SubdirStart > OffsetBase+ExifLength){
                        printf("Illegal GPS directory link");
                    }else{
                        ProcessGpsInfo(SubdirStart, ByteCount, OffsetBase, ExifLength);
                    }
                    continue;
                }
                break;

            case TAG_FOCALLENGTH_35MM:
                // The focal length equivalent 35 mm is a 2.2 tag (defined as of April 2002)
                // if its present, use it to compute equivalent focal length instead of 
                // computing it from sensor geometry and actual focal length.
                ImageInfo.FocalLength35mmEquiv = (unsigned)ConvertAnyFormat(ValuePtr, Format);
                break;
        }
    }


    {
        // In addition to linking to subdirectories via exif tags, 
        // there's also a potential link to another directory at the end of each
        // directory.  this has got to be the result of a comitee!
        unsigned char * SubdirStart;
        unsigned Offset;

        if (DIR_ENTRY_ADDR(DirStart, NumDirEntries) + 4 <= OffsetBase+ExifLength){
            Offset = Get32u(DirStart+2+12*NumDirEntries);
            if (Offset){
                SubdirStart = OffsetBase + Offset;
                if (SubdirStart > OffsetBase+ExifLength || SubdirStart < OffsetBase){
                    if (SubdirStart > OffsetBase && SubdirStart < OffsetBase+ExifLength+20){
                        // Jhead 1.3 or earlier would crop the whole directory!
                        // As Jhead produces this form of format incorrectness, 
                        // I'll just let it pass silently
                        if (ShowTags) printf("Thumbnail removed with Jhead 1.3 or earlier\n");
                    }else{
                        printf("Illegal subdirectory link");
                    }
                }else{
                    if (SubdirStart <= OffsetBase+ExifLength){
                        if (ShowTags) printf("%s    Continued directory ",IndentString);
                        ProcessExifDir(SubdirStart, OffsetBase, ExifLength, NestingLevel+1);
                    }
                }
                if (Offset > ImageInfo.LargestExifOffset){
                    ImageInfo.LargestExifOffset = Offset;
                }
            }
        }else{
            // The exif header ends before the last next directory pointer.
        }
    }

    if (ThumbnailOffset){
        ImageInfo.ThumbnailAtEnd = FALSE;

        if (DumpExifMap){
            printf("Map: %05d-%05d: Thumbnail\n",ThumbnailOffset, ThumbnailOffset+ThumbnailSize);
        }

        if (ThumbnailOffset <= ExifLength){
            if (ThumbnailSize > ExifLength-ThumbnailOffset){
                // If thumbnail extends past exif header, only save the part that
                // actually exists.  Canon's EOS viewer utility will do this - the
                // thumbnail extracts ok with this hack.
                ThumbnailSize = ExifLength-ThumbnailOffset;
                if (ShowTags) printf("Thumbnail incorrectly placed in header\n");

            }
            // The thumbnail pointer appears to be valid.  Store it.
            ImageInfo.ThumbnailOffset = ThumbnailOffset;
            ImageInfo.ThumbnailSize = ThumbnailSize;

            if (ShowTags){
                printf("Thumbnail size: %d bytes\n",ThumbnailSize);
            }
        }
    }
}

//--------------------------------------------------------------------------
// Process a EXIF marker
// Describes all the drivel that most digital cameras include...
//--------------------------------------------------------------------------
void process_EXIF (unsigned char * ExifSection, unsigned int length)
{
    int FirstOffset;

    //FocalplaneXRes = 0;
    //FocalplaneUnits = 0;
    ExifImageWidth = 0;
    OrientationPtr = NULL;


    if (ShowTags){
        printf("Exif header %d bytes long\n",length);
    }

    {   // Check the EXIF header component
        static uchar ExifHeader[] = "Exif\0\0";
        if (memcmp(ExifSection+2, ExifHeader,6)){
            printf("Incorrect Exif header");
            return;
        }
    }

    if (memcmp(ExifSection+8,"II",2) == 0){
        if (ShowTags) printf("Exif section in Intel order\n");
        MotorolaOrder = 0;
    }else{
        if (memcmp(ExifSection+8,"MM",2) == 0){
            if (ShowTags) printf("Exif section in Motorola order\n");
            MotorolaOrder = 1;
        }else{
            printf("Invalid Exif alignment marker.");
            return;
        }
    }

    // Check the next value for correctness.
    if (Get16u(ExifSection+10) != 0x2a){
        printf("Invalid Exif start (1)");
        return;
    }

    FirstOffset = Get32u(ExifSection+12);
    //if (FirstOffset < 8 || FirstOffset > 16){
        // I used to ensure this was set to 8 (website I used indicated its 8)
        // but PENTAX Optio 230 has it set differently, and uses it as offset. (Sept 11 2002)
        //printf("Suspicious offset of first IFD value");
    //}
	if (FirstOffset != 8){
        // I used to ensure this was set to 8 (website I used indicated its 8)
        // but PENTAX Optio 230 has it set differently, and uses it as offset. (Sept 11 2002)

		// PHOTO_MANAGER - put back to ensure it is 8, I will take the chance that a PENTAX Optio 230 is not used
		// and at the same time filter out any junk EXIF data
        printf("Suspicious offset of first IFD value");
		return;
    }

    DirWithThumbnailPtrs = NULL;

    // First directory starts 16 bytes in.  All offset are relative to 8 bytes in.
    ProcessExifDir(ExifSection+8+FirstOffset, ExifSection+8, length-8, 0);

    ImageInfo.ThumbnailAtEnd = ImageInfo.ThumbnailOffset >= ImageInfo.LargestExifOffset ? TRUE : FALSE;

    if (DumpExifMap){
        unsigned a,b;
        printf("Map: %05d- End of exif\n",length-8);
        for (a=0;a<length-8;a+= 10){
            printf("Map: %05d ",a);
            for (b=0;b<10;b++) printf(" %02x",*(ExifSection+8+a+b));
            printf("\n");
        }
    }


    // Compute the CCD width, in millimeters.
    /*if (FocalplaneXRes != 0){
        // Note: With some cameras, its not possible to compute this correctly because
        // they don't adjust the indicated focal plane resolution units when using less
        // than maximum resolution, so the CCDWidth value comes out too small.  Nothing
        // that Jhad can do about it - its a camera problem.
        ImageInfo.CCDWidth = (ExifImageWidth * FocalplaneUnits / FocalplaneXRes);

        if (ImageInfo.FocalLength && ImageInfo.FocalLength35mmEquiv == 0){
            // Compute 35 mm equivalent focal length based on sensor geometry if we haven't
            // already got it explicitly from a tag.
            ImageInfo.FocalLength35mmEquiv = (ImageInfo.FocalLength / ImageInfo.CCDWidth * 36 + 0.5);
        }
    }*/
}

//--------------------------------------------------------------------------
// Cler the rotation tag in the exif header to 1.
// Note: The thumbnail is NOT rotated.  That would be really hard, especially
// stuffing it back into the exif header, because its size might change.
//--------------------------------------------------------------------------
const char * ClearOrientation(void)
{
    if (OrientationPtr == NULL) return NULL;
    
    switch(OrientationNumFormat){
        case FMT_SBYTE:
        case FMT_BYTE:      
            *(uchar *)OrientationPtr = 1;
            break;

        case FMT_USHORT:    
            Put16u(OrientationPtr, 1);                
            break;

        case FMT_ULONG:     
        case FMT_SLONG:     
            memset(OrientationPtr, 0, 4);
            // Can't be bothered to write  generic Put32 if I only use it once.
            if (MotorolaOrder){
                ((uchar *)OrientationPtr)[3] = 1;
            }else{
                ((uchar *)OrientationPtr)[0] = 1;
            }
            break;

        default:
            return NULL;
    }

    return OrientTab[ImageInfo.Orientation];
}
	
long getDecimalPart(int numerator, int denominator)
{
	long integer_part = numerator / denominator;
	long decimal_part;

	decimal_part = ((numerator * 1000000) / denominator);
	decimal_part = decimal_part - (integer_part * 1000000);

	return decimal_part;
}

void formatDecimalForDisplay(char *format_string, long integer_part, long *decimal_part)
{
	if (*decimal_part % 100000 == 0) {
		*decimal_part = *decimal_part / 100000;
		TAP_SPrint(format_string, "%d.%d", integer_part, *decimal_part);
	}
	ef (*decimal_part % 10000 == 0) {
		*decimal_part = *decimal_part / 10000;
		TAP_SPrint(format_string, "%d.%02d", integer_part, *decimal_part);
	}
	ef (*decimal_part % 1000 == 0) {
		*decimal_part = *decimal_part / 1000;
		TAP_SPrint(format_string, "%d.%03d", integer_part, *decimal_part);
	}
	ef (*decimal_part % 100 == 0) {
		*decimal_part = *decimal_part / 100;
		TAP_SPrint(format_string, "%d.%04d", integer_part, *decimal_part);
	}
	ef (*decimal_part % 10 == 0) {
		*decimal_part = *decimal_part / 10;
		TAP_SPrint(format_string, "%d.%05d", integer_part, *decimal_part);
	}
	else {
		TAP_SPrint(format_string, "%d.%06d", integer_part, *decimal_part);
	}
}
//--------------------------------------------------------------------------
// Show the collected image info, displaying camera F-stop and shutter speed
// in a consistent and legible fashion.
//--------------------------------------------------------------------------
void ShowImageInfo()
{
	char buffer[50], formatted_decimal[20];
	no_of_details = 0;

    if (ImageInfo.CameraMake[0]){
		// camera make and model
        //printf("Camera make  : %s\n",ImageInfo.CameraMake);
        //printf("Camera model : %s\n",ImageInfo.CameraModel);
		strcpy(photo_exif[no_of_details].heading, camera_make_t);
		strcpy(photo_exif[no_of_details].detail, ImageInfo.CameraMake);
		no_of_details++;
		strcpy(photo_exif[no_of_details].heading, camera_model_t);
		strcpy(photo_exif[no_of_details].detail, ImageInfo.CameraModel);
		no_of_details++;
    }

    if (ImageInfo.DateTime[0]){

		// convert the date from ccyy:mm:dd to english stylee dd:mm:ccyy
        //printf("Date/Time    : %s\n",ImageInfo.DateTime);
		strcpy(photo_exif[no_of_details].heading, date_time_t);
		strcpy(buffer, "");
		strncat(buffer, ImageInfo.DateTime+8, 2);
		strcat(buffer, "/");
		strncat(buffer, ImageInfo.DateTime+5, 2);
		strcat(buffer, "/");
		strncat(buffer, ImageInfo.DateTime, 4);
		strcat(buffer, " ");
		strncat(buffer, ImageInfo.DateTime+11, 5);
		strcpy(photo_exif[no_of_details].detail, buffer);
		no_of_details++;
    }

	// width x height
    //printf("Resolution   : %d x %d\n",ImageInfo.Width, ImageInfo.Height);
	strcpy(photo_exif[no_of_details].heading, resolution_t);
	TAP_SPrint(buffer, "%d x %d", ImageInfo.Width, ImageInfo.Height);
	strcpy(photo_exif[no_of_details].detail, buffer);
	no_of_details++;

    if (ImageInfo.Orientation > 1){
        // Only print orientation if one was supplied, and if its not 1 (normal orientation)
        //printf("Orientation  : %s\n", OrientTab[ImageInfo.Orientation]);
		strcpy(photo_exif[no_of_details].heading, orientation_t);
		strcpy(photo_exif[no_of_details].detail, OrientTab[ImageInfo.Orientation]);
		no_of_details++;
    }

    if (ImageInfo.IsColor == 0){
		// only display colour info if it is black and white
        //printf("Color/bw     : Black and white\n");
		strcpy(photo_exif[no_of_details].heading, colour_t);
		strcpy(photo_exif[no_of_details].detail, "Black and white");
		no_of_details++;
    }

    if (ImageInfo.FlashUsed >= 0){
		// info on whether the flash was used and its mode
        if (ImageInfo.FlashUsed & 1){    
            //printf("Flash used   : Yes");
			strcpy(photo_exif[no_of_details].heading, flash_used_t);
			strcpy(photo_exif[no_of_details].detail, "Yes");
			no_of_details++;
            switch (ImageInfo.FlashUsed){
	            case 0x5: strcat(photo_exif[no_of_details].detail, " (Strobe light not detected)"); //printf(" (Strobe light not detected)"); 
					break;
	            case 0x7: strcat(photo_exif[no_of_details].detail, " (Strobe light detected)"); //printf(" (Strobe light detected) "); 
					break;
	            case 0x9: strcat(photo_exif[no_of_details].detail, " (manual)"); //printf(" (manual)"); 
					break;
	            case 0xd: strcat(photo_exif[no_of_details].detail, " (manual, return light not detected)"); //printf(" (manual, return light not detected)"); 
					break;
	            case 0xf: strcat(photo_exif[no_of_details].detail, " (maual, return light  detected)"); //printf(" (maual, return light  detected)"); 
					break;
	            case 0x19:strcat(photo_exif[no_of_details].detail, " (auto)"); //printf(" (auto)"); 
					break;
	            case 0x1d:strcat(photo_exif[no_of_details].detail, " (auto, return light not detected)"); //printf(" (auto, return light not detected)"); 
					break;
	            case 0x1f:strcat(photo_exif[no_of_details].detail, " (auto, return light detected)"); //printf(" (auto, return light detected)"); 
					break;
	            case 0x41:strcat(photo_exif[no_of_details].detail, " (red eye reduction mode)"); //printf(" (red eye reduction mode)"); 
					break;
	            case 0x45:strcat(photo_exif[no_of_details].detail, " (red eye reduction mode return light not detected)"); //printf(" (red eye reduction mode return light not detected)"); 
					break;
	            case 0x47:strcat(photo_exif[no_of_details].detail, " (red eye reduction mode return light  detected)"); //printf(" (red eye reduction mode return light  detected)"); 
					break;
	            case 0x49:strcat(photo_exif[no_of_details].detail, " (manual, red eye reduction mode)"); //printf(" (manual, red eye reduction mode)"); 
					break;
	            case 0x4d:strcat(photo_exif[no_of_details].detail, " (manual, red eye reduction mode, return light not detected)"); //printf(" (manual, red eye reduction mode, return light not detected)"); 
					break;
	            case 0x4f:strcat(photo_exif[no_of_details].detail, " (red eye reduction mode, red eye reduction mode, return light detected)"); //printf(" (red eye reduction mode, red eye reduction mode, return light detected)"); 
					break;
	            case 0x59:strcat(photo_exif[no_of_details].detail, " (auto, red eye reduction mode)"); //printf(" (auto, red eye reduction mode)"); 
					break;
	            case 0x5d:strcat(photo_exif[no_of_details].detail, " (auto, red eye reduction mode, return light not detected)"); //printf(" (auto, red eye reduction mode, return light not detected)"); 
					break;
	            case 0x5f:strcat(photo_exif[no_of_details].detail, " (auto, red eye reduction mode, return light detected)"); //printf(" (auto, red eye reduction mode, return light detected)"); 
					break;
            }
        }else{
            //printf("Flash used: No");
			strcpy(photo_exif[no_of_details].heading, flash_used_t);
			strcpy(photo_exif[no_of_details].detail, "No");
			no_of_details++;
            switch (ImageInfo.FlashUsed){
				case 0x18: 
					strcat(photo_exif[no_of_details].detail, " (auto)"); 
					//printf(" (auto)"); 
					break;
            }
        }
        //printf("\n");
    }

    if (ImageInfo.FocalLength_Num && ImageInfo.FocalLength_Den){
		// the distance of the focal length in mm
		long integer_part = ImageInfo.FocalLength_Num / ImageInfo.FocalLength_Den;
		long decimal_part = getDecimalPart(ImageInfo.FocalLength_Num, ImageInfo.FocalLength_Den);
        //printf("Focal length: %d.%dmm\n", (int)integer_part, (int)decimal_part);
		strcpy(photo_exif[no_of_details].heading, focal_length_t);
		formatDecimalForDisplay(formatted_decimal, integer_part, &decimal_part);
		TAP_SPrint(buffer, "%smm", formatted_decimal);
		strcpy(photo_exif[no_of_details].detail, buffer);
		no_of_details++;
    }

    if (ImageInfo.DigitalZoomRatio_Den > 1){
		// Digital zoom used.  Shame on you!
		long integer_part = ImageInfo.DigitalZoomRatio_Num / ImageInfo.DigitalZoomRatio_Den;
		long decimal_part = getDecimalPart(ImageInfo.DigitalZoomRatio_Num, ImageInfo.DigitalZoomRatio_Den);
        //printf("Digital Zoom: %d.%dx\n", integer_part, decimal_part);
		strcpy(photo_exif[no_of_details].heading, digital_zoom_t);
		formatDecimalForDisplay(formatted_decimal, integer_part, &decimal_part);
		TAP_SPrint(buffer, "%sx", formatted_decimal);
		strcpy(photo_exif[no_of_details].detail, buffer);
		no_of_details++;
    }

    //if (ImageInfo.CCDWidth){
        //printf("CCD width    : %4.2fmm\n",(double)ImageInfo.CCDWidth);
		//strcpy(photo_exif[no_of_details].heading, "CCD width");
		//TAP_SPrint(buffer, "%4.2fmm", (double)ImageInfo.CCDWidth);
		//strcpy(photo_exif[no_of_details].detail, buffer);
		//no_of_details++;
    //}

    if (ImageInfo.ExposureTime_Num && ImageInfo.ExposureTime_Den){
		// time taken to expose the photo in secs
		long integer_part = ImageInfo.ExposureTime_Num / ImageInfo.ExposureTime_Den;
		long decimal_part = getDecimalPart(ImageInfo.ExposureTime_Num, ImageInfo.ExposureTime_Den);

        //printf("Exposure time: %d.%d secs", integer_part, decimal_part);
		strcpy(photo_exif[no_of_details].heading, exposure_time_t);
		formatDecimalForDisplay(formatted_decimal, integer_part, &decimal_part);
		TAP_SPrint(buffer, "%s secs", formatted_decimal);
		strcpy(photo_exif[no_of_details].detail, buffer);

        //if (ImageInfo.ExposureTime <= 0.5){
		if ((ImageInfo.ExposureTime_Num * 100) / ImageInfo.ExposureTime_Den <= 50){
            //printf(" (1/%d)", (/*0.5*/ + (1*1000000) / ((ImageInfo.ExposureTime_Num * 1000000) / ImageInfo.ExposureTime_Den)));
			TAP_SPrint(buffer, " (1/%d)", (/*0.5*/1 + (1*1000000) / ((ImageInfo.ExposureTime_Num * 1000000) / ImageInfo.ExposureTime_Den)));
			strcat(photo_exif[no_of_details].detail, buffer);
        }
		//printf("\n");
		no_of_details++;
    }

    if (ImageInfo.ApertureFNumber_Num && ImageInfo.ApertureFNumber_Den){
		// aperture number display (no idea what this is, but format is f/{number})
		long integer_part = ImageInfo.ApertureFNumber_Num / ImageInfo.ApertureFNumber_Den;
		long decimal_part = getDecimalPart(ImageInfo.ApertureFNumber_Num, ImageInfo.ApertureFNumber_Den);
		//printf("Aperture: f/%d.%d\n", integer_part, decimal_part);
		strcpy(photo_exif[no_of_details].heading, aperture_t);
		formatDecimalForDisplay(formatted_decimal, integer_part, &decimal_part);
		TAP_SPrint(buffer, "f/%s", formatted_decimal);
		strcpy(photo_exif[no_of_details].detail, buffer);
		no_of_details++;
    }

    if (ImageInfo.Distance_Num && ImageInfo.Distance_Den){
		// distance of point which is focused in metres
        if (ImageInfo.Distance_Den < 0){
            //printf("Focus distance: Infinite\n");
			strcpy(photo_exif[no_of_details].heading, focus_distance_t);
			strcpy(photo_exif[no_of_details].detail, "Infinite");
			no_of_details++;
        }else{
			long integer_part = ImageInfo.Distance_Num / ImageInfo.Distance_Den;
			long decimal_part = getDecimalPart(ImageInfo.Distance_Num, ImageInfo.Distance_Den);
			//printf("Focus distance: %d.%dm\n", integer_part, decimal_part);
			strcpy(photo_exif[no_of_details].heading, focus_distance_t);
			formatDecimalForDisplay(formatted_decimal, integer_part, &decimal_part);
			TAP_SPrint(buffer, "%sm", formatted_decimal);
			strcpy(photo_exif[no_of_details].detail, buffer);
			no_of_details++;
        }
    }

    if (ImageInfo.ISOequivalent){
		// ISO equivalend (no idea what this is)
        //printf("ISO equiv.   : %d\n", ImageInfo.ISOequivalent);
		strcpy(photo_exif[no_of_details].heading, iso_equivalent_t);
		TAP_SPrint(buffer, "%d", ImageInfo.ISOequivalent);
		strcpy(photo_exif[no_of_details].detail, buffer);
		no_of_details++;
    }

    if (ImageInfo.ExposureBias_Den && ImageInfo.ExposureBias_Num){
        // If exposure bias was specified, but set to zero, presumably its no bias at all,
        // so only show it if its nonzero.
		long integer_part = ImageInfo.ExposureBias_Num / ImageInfo.ExposureBias_Den;
		long decimal_part = getDecimalPart(ImageInfo.ExposureBias_Num, ImageInfo.ExposureBias_Den);
		//printf("Exposure bias: %d.%dm\n", integer_part, decimal_part);
		strcpy(photo_exif[no_of_details].heading, exposure_bias_t);
		formatDecimalForDisplay(formatted_decimal, integer_part, &decimal_part);
		TAP_SPrint(buffer, "%s", formatted_decimal);
		strcpy(photo_exif[no_of_details].detail, buffer);
		no_of_details++;
    }
        
    switch(ImageInfo.Whitebalance) {
		// whitebalance details
        case 1:
            //printf("Whitebalance : Manual\n");
			strcpy(photo_exif[no_of_details].heading, white_balance_t);
			strcpy(photo_exif[no_of_details].detail, "Manual");
			no_of_details++;
            break;
        case 0:
            //printf("Whitebalance : Auto\n");
			strcpy(photo_exif[no_of_details].heading, white_balance_t);
			strcpy(photo_exif[no_of_details].detail, "Auto");
			no_of_details++;
            break;
    }

    //Quercus: 17-1-2004 Added LightSource, some cams return this, whitebalance or both
    switch(ImageInfo.LightSource) {
		// light source setting
        case 1:
			strcpy(photo_exif[no_of_details].heading, "Light Source:");
			strcpy(photo_exif[no_of_details].detail, "Daylight");
			no_of_details++;
            //printf("Light Source : Daylight\n");
            break;
        case 2:
			strcpy(photo_exif[no_of_details].heading, "Light Source:");
			strcpy(photo_exif[no_of_details].detail, "Fluorescent");
			no_of_details++;
            //printf("Light Source : Fluorescent\n");
            break;
        case 3:
			strcpy(photo_exif[no_of_details].heading, "Light Source:");
			strcpy(photo_exif[no_of_details].detail, "Incandescent");
			no_of_details++;
            //printf("Light Source : Incandescent\n");
            break;
        case 4:
			strcpy(photo_exif[no_of_details].heading, "Light Source:");
			strcpy(photo_exif[no_of_details].detail, "Flash");
			no_of_details++;
            //printf("Light Source : Flash\n");
            break;
        case 9:
			strcpy(photo_exif[no_of_details].heading, "Light Source:");
			strcpy(photo_exif[no_of_details].detail, "Fine weather");
			no_of_details++;
            //printf("Light Source : Fine weather\n");
            break;
        case 11:
			strcpy(photo_exif[no_of_details].heading, "Light Source:");
			strcpy(photo_exif[no_of_details].detail, "Shade");
			no_of_details++;
            //printf("Light Source : Shade\n");
            break;
        default:; //Quercus: 17-1-2004 There are many more modes for this, check Exif2.2 specs
            // If it just says 'unknown' or we don't know it, then
            // don't bother showing it - it doesn't add any useful information.
    }

    if (ImageInfo.MeteringMode){ // 05-jan-2001 vcs
		// metering mode (no idea what this is)
        switch(ImageInfo.MeteringMode) {
        case 2:
            //printf("Metering Mode: center weight\n");
			strcpy(photo_exif[no_of_details].heading, metering_mode_t);
			strcpy(photo_exif[no_of_details].detail, "Center weight");
			no_of_details++;
            break;
        case 3:
            //printf("Metering Mode: spot\n");
			strcpy(photo_exif[no_of_details].heading, metering_mode_t);
			strcpy(photo_exif[no_of_details].detail, "Spot");
			no_of_details++;
            break;
        case 5:
            //printf("Metering Mode: matrix\n");
			strcpy(photo_exif[no_of_details].heading, metering_mode_t);
			strcpy(photo_exif[no_of_details].detail, "Matrix");
			no_of_details++;
            break;
        }
    }

    if (ImageInfo.ExposureProgram){ // 05-jan-2001 vcs
        switch(ImageInfo.ExposureProgram) {
        case 1:
            //printf("Exposure     : Manual\n");
			strcpy(photo_exif[no_of_details].heading, exposure_t);
			strcpy(photo_exif[no_of_details].detail, "Manual");
			no_of_details++;
            break;
        case 2:
            //printf("Exposure     : Program (auto)\n");
			strcpy(photo_exif[no_of_details].heading, exposure_t);
			strcpy(photo_exif[no_of_details].detail, "Program (auto)");
			no_of_details++;
            break;
        case 3:
            //printf("Exposure     : Aperture priority (semi-auto)\n");
			strcpy(photo_exif[no_of_details].heading, exposure_t);
			strcpy(photo_exif[no_of_details].detail, "Aperture priority (semi-auto)");
			no_of_details++;
            break;
        case 4:
            //printf("Exposure     : Shutter priority (semi-auto)\n");
			strcpy(photo_exif[no_of_details].heading, exposure_t);
			strcpy(photo_exif[no_of_details].detail, "Shutter priority (semi-auto)");
			no_of_details++;
            break;
        case 5:
            //printf("Exposure     : Creative Program (based towards depth of field\n"); 
			strcpy(photo_exif[no_of_details].heading, exposure_t);
			strcpy(photo_exif[no_of_details].detail, "Creative Program (based towards depth of field)");
			no_of_details++;
            break;
        case 6:
            //printf("Exposure     : Action program (based towards fast shutter speed)\n");
			strcpy(photo_exif[no_of_details].heading, exposure_t);
			strcpy(photo_exif[no_of_details].detail, "Action program (based towards fast shutter speed)");
			no_of_details++;
            break;
        case 7:
            //printf("Exposure     : Portrait Mode\n");
			strcpy(photo_exif[no_of_details].heading, exposure_t);
			strcpy(photo_exif[no_of_details].detail, "Portrait Mode");
			no_of_details++;
            break;
        case 8:
            //printf("Exposure     : Landscape Mode \n");
			strcpy(photo_exif[no_of_details].heading, exposure_t);
			strcpy(photo_exif[no_of_details].detail, "Landscape Mode");
			no_of_details++;
            break;
        default:
            break;
        }
    }
    switch(ImageInfo.ExposureMode){
        case 0: // Automatic (not worth cluttering up output for)
            break;
        case 1: //printf("Exposure Mode: Manual\n");
				strcpy(photo_exif[no_of_details].heading, exposure_t);
				strcpy(photo_exif[no_of_details].detail, "Manual");
				no_of_details++; break;
        case 2: //printf("Exposure Mode: Auto bracketing\n");
				strcpy(photo_exif[no_of_details].heading, exposure_t);
				strcpy(photo_exif[no_of_details].detail, "Auto bracketing");
				no_of_details++; break;
    }

	if (ImageInfo.Process == PROGRESSIVE) {
		strcpy(photo_exif[no_of_details].heading, encoding_t);
		strcpy(photo_exif[no_of_details].detail, progressive_t);
		no_of_details++;
	}

    /*if (ImageInfo.Process != M_SOF0){
        // don't show it if its the plain old boring 'baseline' process, but do
        // show it if its something else, like 'progressive' (used on web sometimes)
        int a;
        for (a=0;;a++){
            if (a >= PROCESS_TABLE_SIZE){
                // ran off the end of the table.
                printf("Jpeg process : Unknown\n");
                break;
            }
            if (ProcessTable[a].Tag == ImageInfo.Process){
                printf("Jpeg process : %s\n",ProcessTable[a].Desc);
                break;
            }
	    }
    }*/

    if (ImageInfo.GpsInfoPresent){
        //printf("GPS Latitude : %s\n",ImageInfo.GpsLat);
        //printf("GPS Longitude: %s\n",ImageInfo.GpsLong);
		strcpy(photo_exif[no_of_details].heading, "GPS Latitude:");
		strcpy(photo_exif[no_of_details].detail, ImageInfo.GpsLat);
		no_of_details++;
		strcpy(photo_exif[no_of_details].heading, "GPS Longitude:");
		strcpy(photo_exif[no_of_details].detail, ImageInfo.GpsLong);
		no_of_details++;
        if (ImageInfo.GpsAlt[0]) 
		{
			//printf("GPS Altitude : %s\n",ImageInfo.GpsAlt);
			strcpy(photo_exif[no_of_details].heading, "GPS Altitude:");
			strcpy(photo_exif[no_of_details].detail, ImageInfo.GpsAlt);
			no_of_details++;
		}
    }

    // Print the comment. Print 'Comment:' for each new line of comment.
    /*if (ImageInfo.Comments[0]){
        int a,c;
        printf("Comment      : ");
        for (a=0;a<MAX_COMMENT;a++){
            c = ImageInfo.Comments[a];
            if (c == '\0') break;
            if (c == '\n'){
                // Do not start a new line if the string ends with a carriage return.
                if (ImageInfo.Comments[a+1] != '\0'){
                    printf("\nComment      : ");
                }else{
                    printf("\n");
                }
            }
			//else{
            //    putchar(c);
            //}
        }
        printf("\n");
    }*/
    //printf("\n");
}

